// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <string.h>
#include "view_all.h"
#include "http_server.h"
#include "utils.h"
#include "picio.h"
#include "log.h"

view_all::view_all(configuration_t *const cfg, const std::string & id, const std::string & descr, const int width, const int height, const int gwidth, const int gheight, resize *const r, const std::vector<view_src_t> & sources, const double switch_interval, std::vector<target *> *const targets, std::vector<filter *> *const filters) : view_ss(cfg, id, descr, width, height, false, r, sources, switch_interval, targets, filters), grid_width(gwidth), grid_height(gheight)
{
	ts_list = new uint64_t[grid_width * grid_height]();

	size_t work_len = grid_width * width * grid_height * height * 3;
	work = (uint8_t *)allocate_aligned(work_len);
}

view_all::~view_all()
{
	free(work);
	delete [] ts_list;
}

std::string view_all::get_html(const std::map<std::string, std::string> & pars) const
{
	std::string dim_img;

	if (width != -1)
		dim_img += myformat(" width=%d", width * grid_width);
	if (height != -1)
		dim_img += myformat(" height=%d", height * grid_height);

	return myformat("<html><head><style></style></head><body><img src=\"stream.mjpeg?int=%s&view-proxy\"%s></body></html>", url_escape(id).c_str(), url_escape(dim_img).c_str());
}

bool view_all::get_frame(const encoding_t pe, const int jpeg_quality, uint64_t *ts, int *o_width, int *o_height, uint8_t **frame, size_t *frame_len, const bool use_previous_when_fail)
{
	bool rc = false;
	const int n_sources = std::min(int(sources.size()), grid_width * grid_height);

	uint64_t si = switch_interval * 1000000.0;
	if (switch_interval < 1.0)
		si = 1000000;

	if (latest_switch == 0)
		latest_switch = get_us();

	uint64_t now = get_us();
	if (now - latest_switch >= si) {
		cur_source_index += n_sources;
		cur_source_index %= sources.size();

		latest_switch = now;

		delete [] ts_list;
		ts_list = new uint64_t[grid_width * grid_height]();
	}

	size_t frame_out_len = grid_width * width * grid_height * height * 3;

	if (!prev_frame)
		prev_frame = (uint8_t *)allocate_aligned(frame_out_len);

	*ts = 0;

	std::vector<std::thread *> threads;

	for(int i=0; i<n_sources; i++) {
		threads.push_back(new std::thread([&, i] {
			size_t offset = (cur_source_index + i) % sources.size();

			instance_t *next_inst = nullptr;
			source *next_source = nullptr;
			find_by_id(cfg, sources.at(offset).id, &next_inst, (interface **)&next_source);
//		printf("%d %lu %lu\n", i, ts_list[i], next_source->get_current_ts());

			if (!next_source)
				log(LL_ERR, "source selection %s in view %s is not known", sources.at(offset).id.c_str(), id.c_str());
			else if (uint64_t s_ts = next_source->get_current_ts(); next_source->is_paused() == false && next_source->is_running() == true && (s_ts > ts_list[i] || ts_list[i] == 0)) {
				uint64_t cur_ts = 0;
				int cur_width = 0, cur_height = 0;
				uint8_t *cur_frame_in = nullptr;
				size_t cur_frame_in_len = 0;

				int cur_grid_x = i % grid_width;
				int cur_grid_y = i / grid_width;

				bool cur_rc = next_source -> get_frame(E_RGB, jpeg_quality, &cur_ts, &cur_width, &cur_height, &cur_frame_in, &cur_frame_in_len, use_previous_when_fail);

				if (cur_rc) {
					*ts = std::max(*ts, cur_ts);
					ts_list[i] = cur_ts;

					rc = true;

					if (cur_width != width || cur_height != height) {
						uint8_t *resized = nullptr;
						r -> do_resize(cur_width, cur_height, cur_frame_in, width, height, &resized);

						free(cur_frame_in);

						for(int y=0; y<height; y++) {
							int offset = y * width * grid_width * 3 + width * cur_grid_x  * 3 + height * width * 3 * grid_width * cur_grid_y;
							memcpy(&work[offset], &resized[y * width * 3], width * 3);
						}

						free(resized);
					}
					else {

						for(int y=0; y<height; y++) {
							int offset = y * width * grid_width * 3 + width * cur_grid_x  * 3 + height * width * 3 * grid_width * cur_grid_y;
							memcpy(&work[offset], &cur_frame_in[y * width * 3], width * 3);
						}

						free(cur_frame_in);
					}
				}
			}
			else {
				log(LL_DEBUG, "source selection %s in view %s is not running/paused", sources.at(offset).id.c_str(), id.c_str());
			}
		}));
	}

	for(auto th : threads) {
		th->join();
		delete th;
	}

	*o_width = width * grid_width;
	*o_height = height * grid_height;

	if (rc) {
		uint8_t *frame_out = (uint8_t *)allocate_aligned(frame_out_len);
		memcpy(frame_out, work, frame_out_len);

		if (filters) {
			instance_t *inst = find_instance_by_interface(cfg, this);
			apply_filters(inst, this, filters, prev_frame, frame_out, *ts, *o_width, *o_height);
			memcpy(prev_frame, frame_out, frame_out_len);
		}

		if (pe == E_RGB) {
			*frame = frame_out;
			*frame_len = frame_out_len;
			return true;
		}
		else if (pe == E_JPEG) {
			rc = write_JPEG_memory(get_meta(), *o_width, *o_height, jpeg_quality, frame_out, (char **)frame, frame_len);
			free(frame_out);

			return rc;
		}
	}

	return false;
}
